﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace Delegate.Invoke_Delegate.BeginInvoke
{
    public partial class Form1 : Form
    {
        private delegate void SomeWorkDelegate();
        private SomeWorkDelegate del;

        public Form1()
        {
            InitializeComponent();
            Control.CheckForIllegalCrossThreadCalls = false;
            del = SomeWork;
        }

        private void SomeWork()
        {
            txt_Message.Text += "\r\n----------委托线程：" + Thread.CurrentThread.ManagedThreadId + "---开始工作\r\n";
            //耗时的操作
            Thread.Sleep(2000);
            txt_Message.Text += "\r\n----------委托线程：" + Thread.CurrentThread.ManagedThreadId + "---开始结束\r\n";
        }

        private void btn_General_Click(object sender, EventArgs e)
        {
            txt_Message.Text = "";
            txt_Message.Text += "主线程："+ Thread.CurrentThread.ManagedThreadId + "---开始工作\r\n";
            //委托方法，在调用委托的线程中执行，本例中就是主线程（UI线程）。
            //执行一些耗时的操作，就会阻塞主线程（UI线程）
            //委托的普通调用就等于方法的直接调用,del();等价于SomeWork();
            del();
            //SomeWork();
            txt_Message.Text += "\r\n主线程：" + Thread.CurrentThread.ManagedThreadId + "---开始结束\r\n";
        }

        private void btn_Main_Invoke_Click(object sender, EventArgs e)
        {
            txt_Message.Text = "";
            txt_Message.Text += "主线程：" + Thread.CurrentThread.ManagedThreadId + "---开始工作\r\n";
            //委托的同步调用，其实就是等价于委托的普通调用。
            del.Invoke();
            txt_Message.Text += "\r\n主线程：" + Thread.CurrentThread.ManagedThreadId + "---开始结束\r\n";
        }

        private void btn_Main_BeginInvoke_Click(object sender, EventArgs e)
        {
            txt_Message.Text = "";
            txt_Message.Text += "主线程：" + Thread.CurrentThread.ManagedThreadId + "---开始工作\r\n";
            //委托异步调用
            //1、委托方法，在线程池中分配的子线程中执行。
            //2、主线程和子线程同时执行。
            //3、子线程结束之后，如果有返回值得话，将返回值传递给主线程。如果有回调函数的话，继续在子线程中执行回调函数。

            //有异常，控件不能在子线程中访问修改。
            //避免这类异常有两种方法   
            //1、手动关闭控件的跨线程安全检查Control.CheckForIllegalCrossThreadCalls = false;(不建议使用)
            //2、使用控件的Invoke方法。（推荐使用）
            del.BeginInvoke(null,null);
            txt_Message.Text += "\r\n主线程：" + Thread.CurrentThread.ManagedThreadId + "---开始结束\r\n";
        }

        private void btn_Sub_Invoke_Click(object sender, EventArgs e)
        {
            txt_Message.Text = "";
            txt_Message.Text += "主线程：" + Thread.CurrentThread.ManagedThreadId + "---开始工作\r\n";

            //开启新的线程执行委托，主线程（UI线程）继续向下执行
            new Thread(() => {
                txt_Message.Text += "\r\n----子线程：" + Thread.CurrentThread.ManagedThreadId + "---开始工作\r\n";
                //委托在调用线程中执行，并阻塞调用线程，知道委托方法执行结束。
                del.Invoke();
                txt_Message.Text += "\r\n----子线程：" + Thread.CurrentThread.ManagedThreadId + "---开始结束\r\n";
            }).Start();

            txt_Message.Text += "\r\n主线程：" + Thread.CurrentThread.ManagedThreadId + "---开始结束\r\n";
        }

        private void btn_Sub_BeginInvoke_Click(object sender, EventArgs e)
        {
            txt_Message.Text = "";
            txt_Message.Text += "主线程：" + Thread.CurrentThread.ManagedThreadId + "---开始工作\r\n";

            //开启新的线程执行委托，主线程（UI线程）继续向下执行
            new Thread(() =>
            {
                txt_Message.Text += "\r\n----子线程：" + Thread.CurrentThread.ManagedThreadId + "---开始工作\r\n";
                //在线程池中分配的子线程中执行委托方法，调用委托的线程继续向下执行。
                del.BeginInvoke(null, null); 
                txt_Message.Text += "\r\n----子线程：" + Thread.CurrentThread.ManagedThreadId + "---开始结束\r\n";
            }).Start();

            txt_Message.Text += "\r\n主线程：" + Thread.CurrentThread.ManagedThreadId + "---开始结束\r\n";
        }
    }
}
